"""Parallel processing functions for distill and inference pipelines."""

import os
from inspect import getsource
from pathlib import Path
from typing import Callable, Tuple

import pandas as pd
from tqdm import trange

from zeno.api import DistillReturn, ModelReturn, ZenoOptions
from zeno.classes.base import ZenoColumn, ZenoColumnType


def predistill_data(
    fn: Callable[[pd.DataFrame, ZenoOptions], DistillReturn],
    column: ZenoColumn,
    options: ZenoOptions,
    cache_path: str,
    df: pd.DataFrame,
    batch_size: int,
    pos: int,
) -> Tuple[ZenoColumn, pd.Series]:
    col_hash = str(column)

    # To prevent SettingWithCopyWarning
    col = df[col_hash].copy()

    save_path = Path(cache_path, col_hash + ".pickle")
    to_predict_indices = col.loc[pd.isna(col)].index

    if len(to_predict_indices) > 0:
        if len(to_predict_indices) < batch_size:
            out = fn(df.loc[to_predict_indices], options)
            col.loc[to_predict_indices] = out.distill_output
            col.to_pickle(str(save_path))
        else:
            for i in trange(
                0,
                len(to_predict_indices),
                batch_size,
                desc="preprocessing " + fn.__name__,
                position=pos,
            ):
                out = fn(df.loc[to_predict_indices[i : i + batch_size]], options)
                col.loc[to_predict_indices[i : i + batch_size]] = out.distill_output
                col.to_pickle(str(save_path))
    return (column, col)


def run_inference(
    fn: Callable[[str], Callable[[pd.DataFrame, ZenoOptions], ModelReturn]],
    options: ZenoOptions,
    model_path: str,
    cache_path: str,
    df: pd.DataFrame,
    batch_size: int,
    pos: int,
) -> Tuple[ZenoColumn, ZenoColumn, pd.Series, pd.Series]:
    model_name = os.path.basename(model_path).split(".")[0]

    model_col_obj = ZenoColumn(
        column_type=ZenoColumnType.OUTPUT,
        name=model_name,
    )
    embedding_col_obj = ZenoColumn(
        column_type=ZenoColumnType.EMBEDDING,
        name=model_name,
    )
    model_hash = str(model_col_obj)
    embedding_hash = str(embedding_col_obj)
    model_col = df[model_hash].copy()
    embedding_col = df[embedding_hash].copy()

    model_save_path = Path(cache_path, model_hash + ".pickle")
    embedding_save_path = Path(cache_path, embedding_hash + ".pickle")

    to_predict_indices = model_col.loc[pd.isna(model_col)].index

    if len(to_predict_indices) > 0:
        model_fn = fn(model_path)

        # Make output folder if function uses output_path
        src = getsource(model_fn)
        if "output_path" in src:
            file_cache_path = os.path.join(cache_path, model_hash)
            os.makedirs(file_cache_path, exist_ok=True)
            options = options.copy(update={"output_path": file_cache_path})

        if len(to_predict_indices) < batch_size:
            out = model_fn(df.loc[to_predict_indices], options)

            # Check if we also get embedding
            if out.embedding is not None:
                # Loop for setting since setting with loc and NDArray can cause issues.
                for i, idx in enumerate(to_predict_indices):
                    model_col.at[idx] = out.model_output[i]
                    embedding_col.at[idx] = out.embedding[i]
                embedding_col.to_pickle(str(embedding_save_path))
            else:
                model_col.loc[to_predict_indices] = out.model_output

            model_col.to_pickle(str(model_save_path))

        else:
            for i in trange(
                0,
                len(to_predict_indices),
                batch_size,
                desc="Inference on " + model_name,
                position=pos,
            ):
                out = model_fn(df.loc[to_predict_indices[i : i + batch_size]], options)

                # Check if we also get embedding
                if out.embedding is not None:
                    for i, idx in enumerate(to_predict_indices[i : i + batch_size]):
                        model_col.at[idx] = out.model_output[i]
                        embedding_col.at[idx] = out.embedding[i]

                    embedding_col.to_pickle(str(embedding_save_path))
                else:
                    model_col.loc[
                        to_predict_indices[i : i + batch_size]
                    ] = out.model_output

                model_col.to_pickle(str(model_save_path))

    return (model_col_obj, embedding_col_obj, model_col, embedding_col)


def postdistill_data(
    fn: Callable[[pd.DataFrame, ZenoOptions], DistillReturn],
    model: str,
    options: ZenoOptions,
    cache_path: str,
    df: pd.DataFrame,
    batch_size: int,
    pos: int,
):
    col_obj = ZenoColumn(
        column_type=ZenoColumnType.POSTDISTILL,
        name=fn.__name__,
        model=model,
    )
    col_hash = str(col_obj)
    col = df[col_hash].copy()

    output_obj = ZenoColumn(
        column_type=ZenoColumnType.OUTPUT,
        name=model,
    )
    output_hash = str(output_obj)

    save_path = Path(cache_path, col_hash + ".pickle")

    to_predict_indices = col.loc[pd.isna(col)].index

    local_options = options.copy(
        update={
            "output_column": output_hash,
            "output_path": os.path.join(cache_path, output_hash),
        }
    )

    if len(to_predict_indices) > 0:
        if len(to_predict_indices) < batch_size:
            out = fn(df.loc[to_predict_indices], local_options)
            col.loc[to_predict_indices] = out.distill_output
            col.to_pickle(str(save_path))
        else:
            for i in trange(
                0,
                len(to_predict_indices),
                batch_size,
                desc="postprocessing " + fn.__name__ + " on " + model,
                position=pos,
            ):
                out = fn(df.loc[to_predict_indices[i : i + batch_size]], local_options)
                col.loc[to_predict_indices[i : i + batch_size]] = out.distill_output
                col.to_pickle(str(save_path))
    return (col_obj, col)



from zeno.api import (
    distill,
    DistillReturn,
    inference,
    InferenceReturn,
    metric,
    MetricReturn,
    model,
    ModelReturn,
    ZenoOptions,
    ZenoParameters,
)
from zeno.runner import zeno

__all__ = [
    "DistillReturn",
    "InferenceReturn",
    "MetricReturn",
    "ModelReturn",
    "ZenoOptions",
    "ZenoParameters",
    "distill",
    "inference",
    "metric",
    "model",
    "zeno",
]



"""External API for Zeno. Includes decorator functions, return types, and options."""

import functools
from typing import Any, Callable, Dict, List, Union

from gradio.interface import IOComponent  # type: ignore
from numpy.typing import NDArray
from pandas import DataFrame, Series
from pydantic import BaseModel


class ZenoOptions(BaseModel):
    """Parameters passed to Zeno test functions.

    Args:
        id_column (str): Column in dataframe with unique identifiers.
        data_column (str): Column in dataframe with either raw data or path to data.
        label_column (str): Column in dataframe with
                            either raw labels or path to labels.
        output_column (str): Column in dataframe with a given
                             model's raw output or path to output
        data_path (str): Path to directory with data files.
        label_path (str): Path to directory with label files.
        output_path (str): Path to directory with a given model's output.
        distill_columns (map[str, str]): Map from distill function name to
        distill column.
    """

    id_column: str
    data_column: str
    label_column: str
    output_column: str
    distill_columns: Dict[str, str]
    data_path: str
    label_path: str
    output_path: str


class ZenoParameters(BaseModel):
    """Options passed to the backend processing pipeline."""

    metadata: DataFrame
    view: str = ""
    functions: List[Callable] = []
    models: List[str] = []
    id_column: str = ""
    data_column: str = ""
    label_column: str = ""
    data_path: str = ""
    label_path: str = ""
    batch_size: int = 1
    cache_path: str = ""
    calculate_histogram_metrics = True
    editable: bool = True
    serve: bool = True
    samples: int = 30
    port: int = 8000
    host: str = "localhost"

    class Config:
        arbitrary_types_allowed = True


class ModelReturn(BaseModel):
    """Return type for model functions.

    Args:
        model_output (Series | List): Model output for each sample.
        embedding (Series | List[List[float]] | List[NDArray] | NDArray | None):
        High-dimensional embedding for each sample. Optional.
        other_returns (Dict[str, Series | List] | None): Other returns from the model
        to be shown as metadata columns in the UI.
    """

    model_output: Union[Series, List[Any]]
    embedding: Union[Series, List[List[float]], List[NDArray], NDArray, None] = None
    other_returns: Union[Dict[str, Union[Series, List[Any]]], None] = None

    class Config:
        arbitrary_types_allowed = True


class DistillReturn(BaseModel):
    """Return type for distill functions

    Args:
        distill_output (Series | List): Distill outputs for each sample.
    """

    distill_output: Union[Series, List[Any]]

    class Config:
        arbitrary_types_allowed = True


class MetricReturn(BaseModel):
    """Return type for metric functions.

    Args:
        metric (float): Average metric over subset of data
        variance (float): Variance of metric over subset of data
    """

    metric: float
    variance: Union[float, None] = None

    class Config:
        arbitrary_types_allowed = True


class InferenceReturn(BaseModel):
    """Return type for inference UI functions.
    Take a look at the Gradio documentation for details.

    Args:
        input_components (List[Blocks]): List of input components.
        output_components (Blocks): Output component.
        input_columns (List[str]): List of input column names matching the
        input components.
    """

    input_components: List[IOComponent]
    output_component: IOComponent
    input_columns: List[str]

    class Config:
        arbitrary_types_allowed = True


def model(func: Callable[[str], Callable[[DataFrame, ZenoOptions], ModelReturn]]):
    """Decorator function for model functions.

    Args:
        func (Callable[[str], Callable[[DataFrame, ZenoOptions], ModelReturn]]):
        A function that that takes a model name and returns a model function, which
        itself returns a function that takes a DataFrame and ZenoOptions and returns a
        ModelReturn.
    """

    @functools.wraps(func)
    def _wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    _wrapper.predict_function = True  # type: ignore
    return _wrapper


def distill(func: Callable[[DataFrame, ZenoOptions], DistillReturn]):
    """Deocrator function for distill functions.

    Args:
        func (Callable[[DataFrame, ZenoOptions], DistillReturn]):
        A function that takes a DataFrame and ZenoOptions and returns a DistillReturn.
    """

    @functools.wraps(func)
    def _wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    _wrapper.distill_function = True  # type: ignore
    return _wrapper


def metric(func: Callable[[DataFrame, ZenoOptions], MetricReturn]):
    """Decorator function for metric functions.

    Args:
        func (Callable[[DataFrame, ZenoOptions], MetricReturn]):
            A metric function that takes a DataFrame and ZenoOptions and
            returns a MetricReturn with an average metric value and optional variance.
    """

    @functools.wraps(func)
    def _wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    _wrapper.metric_function = True  # type: ignore
    return _wrapper


def inference(func: Callable[[ZenoOptions], InferenceReturn]):
    """A decorator function for inference UI using Gradio.

    Args:
        func (Callable[[ZenoOptions], InferenceReturn]): Function that returns a set of
        Gradio components and input columns to scaffold an inference UI.
    """

    @functools.wraps(func)
    def _wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    _wrapper.inference_function = True  # type: ignore
    return _wrapper
